//=============================================================================
// System  : CVMEditor - Virtual Memory Text Editor Library
// File    : CVMEditor.h
// Author  : Eric Woodruff  (Eric_Woodruff@CompuServe.com)
// Updated : Wed 07/08/98 01:00:00
// Note    : Copyright 1994-98, Eric Woodruff, All rights reserved
// Compiler: Visual C++
//
// This is the header file for the Virtual Memory Text Editor classes.
// It defines the CVMEditor class and supporting definitions.
//
// Version    Date    Who   Comments
// ============================================================================
// 1.0.0    06/04/97  EFW   Ported the code to Visual C++
//=============================================================================

#if !defined(__CVMEDITOR_H)
#define __CVMEDITOR_H

#if !defined(__CVMEDITORDEFS_H)
#include "CVMEditorDefs.h"
#endif

#if !defined(__CVMPOINT_H)
#include "CVMPoint.h"
#endif

#if !defined(__CVIRTMEM_H)
#include "CVirtualMemory.h"
#endif

//*****************************************************************************
//
// Class CVMEditor
//
class CVMEditor
{
public:
    // CVirtualMemory is a friend of CVMEditor so that it can update some
    // of the private data members to keep the editor and the buffer
    // in sync.
    friend class CVirtualMemory;

    // Static data used by all instances.
    static CVMEditor *clipboard;            // Clipboard (see docs)

    static CVMEditorDialog editorDialog;    // User interaction dialog
    static CVMSignsOfLife  signsOfLife;     // Signs of Life function

    static const char *szCharLeft;          // Valid Enter Matching
    static const char *szCharRight;         // characters.

    static char chEOLMarker;        // wrapPara() line length inhibitor char

    static BOOL bInsertMode;                // Insert mode toggle

    // Global options structure shared by all instances.
    // Some of these become the defaults for each new instance.
    static GlobalEditorOptions GlobalOpts;

    // Current search/replace criteria (also shared by all instances)
    static USHORT nSearchDirection;
    static USHORT nSearchScope;
    static USHORT nSearchOrigin;
    static char   szFindStr[maxSearchString];
    static char   szReplaceStr[maxSearchString];

    // Other public variables.

    // The virtual memory buffer object.  The virtual memory manager is a
    // separate entity to allow the modification of the underlying virtual
    // memory engine.  Also, multiple editors can share the same buffer
    // object to conserve memory.
    CVirtualMemory *Buffer;

    // Flags: Is valid or not, line has been edited
    BOOL bIsValid, bLineEdited;

    LocalEditorOptions Opts;        // Options local to each instance

    int  nSelecting;                // Type of select mode in use

    CVMPoint delta,                 // First visible row and column on screen
             screenSize;            // Screen size

protected:
    static DWORD dwVMKeyMap[maxKeyMaps];    // Key map resource IDs

    char szText[maxLineLength + 1]; // Current edit line buffer

    CVMPoint curPos;                // Cursor position
    long lLineCnt;                  // Line count

    int  nSelStartC, nSelEndC;      // Block start and end columns for
                                    // XY block selection.
    long lSelStart, lSelEnd,        // Block start and end lines
         lFindLine;                 // Line where search text was found

    SignsOfLife slType;             // Signs of Life status variables.
    long slCount, slReportMark,     // These are used internally to
         slBytes, slTotal;          // track when they should occur.

    long lFlags;                    // Update hint flags

    int  nCurPosx_t;                // Cursor position in edited line

    BOOL bFirstLoad;                // Signals first time file is loaded

    long lLastChg,                  // Last modified line
         lPrevLine;                 // Previous cursor line

// ***************************************************************************
public:
    CVMEditor(int bufSizeInK = 8, const char *pszFName = NULL,
        BOOL bSuppressLoad = FALSE);
    virtual ~CVMEditor();

    // Change key map IDs
    static void changeKeyMap(int nIdx, DWORD nID);

    virtual void update(long lModFlags, BOOL bEnable = TRUE);
    virtual void setXIndex(int nNewPos);
    virtual void setScrIndex(int nNewPos);
    virtual void scrollTo(int x, long y);

    virtual BOOL isWordChar(int ch);

    virtual CVMEditor *checkForExistingEditor(const char *pszFilename);

    virtual void doneBuffer(BOOL releaseMem = TRUE, BOOL reloadFile = FALSE);

    // These provide basic clipboard support throught the static 'clipboard'
    // data member.  Derive a new class to implement support for other
    // clipboard types (i.e. the Windows clipboard).
    virtual BOOL clipCopy(void);
    virtual BOOL clipCut(void);
    virtual BOOL clipPaste(void);

    // Measure text for word wrapping
    virtual LineEndPos MeasureText(const char *pszLine, int nLen, int nCol,
        int *pnWrapCol = NULL);

    // Default command and character handlers
    virtual void processCommand(UINT nCommand, BOOL bEnableSelect);
    virtual void processChar(UINT nChar, BOOL bInsertLiteral = FALSE);

// ***************************************************************************

    // Add or remove a bookmark or set its data pointer.  Virtual
    // to allow special handling of the optional data if needed.
    virtual Bookmark *AddBookmark(int nType, long lLine,
        void (*vfpCallback)(Bookmark *) = NULL, void *vpData = NULL);
    virtual void RemoveBookmark(int nType, long lLine);
    virtual void RemoveBookmark(int nIdx);
    virtual void RemoveBookmark(Bookmark *pBookmark);
    virtual void ModifyBookmark(int nType, long lLine, long lCount);
    virtual void ModifyBookmark(int nIdx, long lCount);
    virtual void SetBookmarkData(int nType, long lLine, void *vpData);

    // Sends notification that bookmark positions changed when lines
    // were added or deleted.
    virtual void BookmarksChanged(void);

    // Clear all bookmarks of a given type
    void ClearBookmarks(int nType = BK_ALLTYPES, BOOL bNotClosing = TRUE);

    // Find a bookmark of the given type at the given line
    Bookmark *FindBookmark(int nType, long lLine);

    // Jump to bookmark and return pointer to it
    Bookmark *JumpToBookmark(int nType, long lLine);

// ***************************************************************************

    // Get resource ID or accelerator table handle for key map
    static inline DWORD GetKeyMapID(int nIdx);

    // Allow the option bits to be set externally
    inline void setOption(long efConst, BOOL bOn);

    // Check undo/redo state
    inline BOOL canUndo(void);

    // Cursor position and limits shouldn't be modified directly so they
    // are protected.  These two inline functions will let you know what
    // they are currently set to.
    inline void cursorPos(int &nCol, long &lLine);
    inline void textLimits(int &nCol, long &lLine);

    // This one is used internally for clipboard operations
    inline int getTextIndex(void);

    // It is more expedient in an iterative, line-by-line operation to call
    // the virtual memory buffer text retrieval functions directly and work
    // with the Text[] line (a Compile operation for example).  This returns
    // a const char * to Text[] for those cases.
    inline const char *lineText(void);

    // Get view update hint flags
    inline long getViewFlags(void);

// ***************************************************************************

    void  setSelect(int fromX, long fromY, int toX, long toY,
              BOOL goTop = FALSE);
    void  putCursorAt(int x, long y, BOOL bCenterCursor = FALSE);

    long  getTextSize(void);

    int   pad(int nPos);
    int   trim(char *szLine, char *pChar);

    void  moveChar(int move);
    void  moveWord(int move);
    void  moveLine(long  move);
    void  setCurrLine(long lLine);

    BOOL  insertChar(char ch, BOOL insChar);
    void  insertTab(BOOL bCenterCursor);
    void  insertCR(void);

    void  deleteChar(BOOL DelRight);
    void  deleteTab(BOOL bCenterCursor);
    void  deleteLine(void);

    BOOL  hasSelection(void);
    void  anchorBlock(CVMPoint &oldCurPos, BOOL isSelecting);
    void  adjustBlock(int difference);

    void  deleteBlock(void);             // Non-clipboard delete, move, copy
    void  moveBlock(void);
    void  copyBlock(void);

    BOOL  insertFrom(CVMEditor *insEd);  // Clipboard copy, cut, paste
    BOOL  insertAtCursor(const char *text);
    void  insertBlock(const char *text, long lSize);

    long  getSelectedTextSize(char LineTerm);
    long  getSelectedText(char *buffer, long length, char LineTerm,
        BOOL *NotDone);

#if !defined(NO_WRAPCODE)
    BOOL wrapLine(void);
    void wrapPara(void);
    void reformatAll(void);
    void doAutoWrap(void);
#else
    #define doAutoWrap()
    #define reformatAll()
#endif

    void alterText(long from, long to, UINT cmnd);

    int  SearchDialog(BOOL Replace);
    int  doSearchReplace(void);
    BOOL search(void);
    void FindMatching(void);

    // Load/save handlers
    virtual void loadFileNow(const char *pszFName = NULL);
    BOOL save(const char *pszFName = NULL);
    BOOL saveAs(void);

    void readFile(const char *pszFName);
    long readText(const char *pszFName, long lLine, BOOL bInitialLoad);
    BOOL writeFile(long lFrom, long lTo, const char *pszFName,
        BOOL bUpdateTimeStamp = FALSE);

    // Get a file's time stamp
    BOOL getFileTimeStamp(PFILETIME pFileTime, const char *pszFilename);

    // Auto save hook
    virtual BOOL autoSave(const char *pszFilename = NULL,
        BOOL bOverwriteOriginal = FALSE);
};

// Inline function bodies
inline DWORD CVMEditor::GetKeyMapID(int nIdx)
    {   return CVMEditor::dwVMKeyMap[nIdx];    }

inline void CVMEditor::setOption(long efConst, BOOL bOn)
    { (bOn) ? (Opts.lOperation |= efConst) : (Opts.lOperation &= ~efConst); }

inline BOOL CVMEditor::canUndo(void)
    {   return BOOL((Opts.lOperation & efCanUndo) != 0L);  }

inline void CVMEditor::cursorPos(int &nCol, long &lLine)
{
    nCol = curPos.x;
    lLine = curPos.y;
}

inline void CVMEditor::textLimits(int &nCol, long &lLine)
{
    nCol = maxLineLength;       // Always maxLineLength
    lLine = lLineCnt;
}

inline int CVMEditor::getTextIndex(void)
    {   return nCurPosx_t;   };

inline const char *CVMEditor::lineText(void)
    {   return szText;    }

inline long CVMEditor::getViewFlags(void)
    {   return lFlags;  }

#endif  // __CVMEDITOR_H
