//=============================================================================
// System  : CVMEditor - Virtual Memory Text Editor Library
// File    : CVirtualMemory.h
// Author  : Eric Woodruff  (Eric_Woodruff@CompuServe.com)
// Updated : Wed 07/08/98 01:00:00
// Note    : Copyright 1994-98, Eric Woodruff, All rights reserved
// Compiler: Visual C++
//
// This header file declares the constants and prototypes for the
// CVirtualMemory class used by the editor classes.
//
// The maximum document size in bytes is PageSize * MaxSlots.  The
// paging system is adaptive and can adjust MaxSlots upward each time
// the limit is hit.  To start with, MaxSlots is set to 512 which allows
// a file of 512K, 1 Meg, or 2 Meg (depending on page size) to be loaded
// before MaxSlots is adjusted.  Theoretically, MaxSlots can be adjusted
// up to 32,767 which would allow for a file of up to 32 Meg, 64 Meg, or
// 128 Meg (depending on page size).  Of course, I've not been able to
// test this limit, so don't expect miracles <g>.  Also, I say
// "depending on page size", because the paging system adjusts the page
// size based on the size of its buffer. Anything under 12K uses a 1K
// page, anything under 20K uses a 2K page, and anything 20K or larger
// will use a 4K page to maximize efficiency.
//
// The line and undo information pages are virtualized along with the
// text so the real maximum will be somewhat less based on the number of
// lines in the file and whether undo is enabled or disabled.  The
// editor is capable of containing more than 32,767 lines because it
// uses type long values for the line numbers.  The overhead for line
// information is about 1K per 256 lines in the file (about 8%-15%).
// Undo information overhead will vary based on whether or not it is
// enabled and how much editing is actually done on the file.  When
// disabled, the undo information is still retained for the purpose of
// recycling the discarded virtual memory positions for other text as it
// is edited.
//
// NOTE:  If compiling the library for debugging, you can add the
//        definition CVME_DIAG to the project settings to enable
//        internal diagnostics.  Specify the name in all uppercase
//        letters if used!
//
// Version    Date    Who   Comments
// ============================================================================
// 1.0.0    06/04/97  EFW   Ported the code to Visual C++
//=============================================================================

#if !defined(__CVIRTMEM_H)
#define __CVIRTMEM_H

#if !defined(__STRING_H)
#include <string.h>
#endif  // __STRING_H

#if !defined(__CBKARRAY_H)
#include "CBkArray.h"
#endif

#define EOS     '\x0'

// ****************************************************************************
// Maximum number of pages in memory using 4096 byte pages.  The actual
// number will be determined by the amount of memory requested for the
// buffer.  If the document is too big to be held in memory then a
// temporary disk file will be created.  Pages in the buffer may have a
// corresponding disk image so there should be enough disk space to hold
// the whole document plus enough space for the line and undo information.

#define RAMSLOTS   15

// Default MaxSlots size and adjustment increment
#define SLOTCOUNT   512

// Virtual memory page size
#define VMEMPAGESIZE    65536L

#define LRU_RECENT  LONG_MAX    // Flags the most recently swapped in page(s)
#define UNASSIGNED  SHRT_MIN    // Flags an unassigned virtual page

// This macro returns the address of the requested page in real memory.
// Page * PageSize may be greater than 32767! MUST BE CAST AS UNSIGNED!!
//
// It is a macro rather than an inline function to eliminate function call
// overhead if the library is compiled using out-of-line inline functions.
#define getSlotAddr(nPage) \
    (PageBuffer + ((unsigned)nPage * (unsigned)nPageSize))

// ****************************************************************************

// This is the structure that holds the header information for each
// page of line number information.  It is composed of the address of
// the next and previous line info pages and the start line for this page
// along with the current number of line structures within the page.
struct LineHeader
{
    short nPrevPage,
          nNextPage;

    long  lFirstLineNo,
          lCount;           // Count's type is long to align the structure
};                          // size for easier calculation of LinesPerVPage.

// This is the structure that holds the address of each line.
// It is composed of a virtual page number and an offset into that page.
struct LineAddress
{
    short nPage,
          nOffset;
};

// Types of undo/redo information
enum UndoType { undoInsert, undoReplace, undoDelete, undoEnd };

// This is the structure that holds the undo/redo data.
// The size parameter is present so that the undo/redo data also serves as
// a free space list when undo/redo is disabled.  Discarded virtual memory
// can then be recycled.
struct UndoData
{
    UndoType Type;      // Type of undo/redo step

    long  lStep,        // Undo/redo step indicator
          lLine;        // Line for the operation

    short nPage,        // Location and size
          nOffset,
          nSize;
};

class CVMEditor;

// This is a simple linked list structure used to keep track of all the
// buffer's owners.  It doesn't need to be anything fancy, just something to
// be used for updating multiple owners when necessary.
// Multiple ownership of the same buffer can be used to reduce memory
// overhead when two or more editors are used to modify the same file.
struct OwnerList
{
    CVMEditor *Owner;
    OwnerList  *nextOwner;
};

// ****************************************************************************

// This is the virtual memory class used for storing the editor text.
class CVirtualMemory
{
private:

    // Maximum number of pages that can be allocated by an instance before
    // forcing disk swapping.  Useful for preventing one editor from
    // consuming all available memory.
    static int nMaxAllocatedPages;

    // The GlobalAlloc() handle for this editor and a pointer to the locked
    // block location.
    HGLOBAL hMemHandle;
    char   *pMemLocation;

    // While memory is still available, this is set to INT_MAX.  When
    // not INT_MAX, it refers to the last page that was allocated in
    // physical memory.  This signals that such memory is no longer
    // available and disk space is being used.  Requests that reference
    // data past the page number in this variable will be read/written
    // from/to a disk swapping file.
    int nLastVMemPage;

    // File handle for the paging file if it is created
    HANDLE hPageFile;

    // Pages allocated so far.  If greater than zero, swapping has
    // been started.
    int nMemPagesAllocated;

    // Pointer to the list of owners
    OwnerList *Owners;

    // Pointer to the currently selected owner of this buffer.  Used to
    // access and update some necessary variables in the owner.
    CVMEditor *owner;

    // Pointer to the LRU page buffer.  Currently, this buffer is
    // never larger than 60K.  This may change in future versions.
    char *PageBuffer;

    // Pointers to the resident line information page data and undo data page
    LineHeader  *hp;
    LineAddress *lp;
    UndoData    *up;

    // These two track the last unmodified point in undo/redo
    int  nLastSavedPage, nLastSavedPoint;

    BOOL bIsValid;                  // Used to signal good initialization

    int  nBufSize,                  // LRU buffer size
         nPageSize,                 // Page size
         nPgPerVMemPage,            // Virtual pages per swap page
         nLinesPerVPage,            // Line structures per virtual page
         nUndosPerVPage,            // Undo structures per virtual page
         nSlotsInRAM,               // Actual number of slots in memory
         nMaxSlots,                 // Total number of allocatable pages
         nEditPage,                 // Page number that receives new text
         nInsPoint,                 // Insertion point for text in EditPage
         nLineInfoPage,             // Page receiving line number information
         nUndoInfoPage,             // Page receiving undo information
         nUndoPoint;                // Current undo step position

    long lLRU_Cntr;                 // Counter for Least Recently Used paging

    long  LRU_Prty[RAMSLOTS];       // Least Recently Used swap priority info
    short SwapSlotNo[RAMSLOTS];     // Swap slot number associated with the
                                    // memory slot

    short *SlotLocn;                // Current page location (RAM/Swap slot #)
    char  *VirtSlotInUse;           // Virtual memory slot usage indicators
    char  *SwapSlotInUse;           // Swap slot usage indicators

    long lTextSize;                 // Total number of characters

public:

    CBkArray Bookmarks;             // Bookmark array

    char szFilename[MAX_PATH];      // Filename (if used)

    // This can be used to record the time stamp of the file that
    // is loaded into the buffer.  It is currently used by the
    // CVM_RESYNCHxxx commands in CVMEditor.
    FILETIME ftTimeStamp;

    // Flag to signal that the text in the buffer has been modified
    BOOL bModified;

    // The editor controls what set of undo steps constitutes a block
    // by incrementing the UndoBlock variable.  As long as UndoBlock has
    // an identical value in prior or subsequent steps, UndoRedo() will
    // loop so as to undo or redo all associated modifications at once.
    // This is mainly for block operations (delete, move, copy, alter
    // character case, text reformatting, etc).
    long lUndoBlock;

    CVirtualMemory(CVMEditor *anOwner, int bufSizeInK);
    ~CVirtualMemory();

// ****************************************************************************

    static void setMaxAllocation(long lMaxSizeInK)
        {   nMaxAllocatedPages = int(lMaxSizeInK * 1024L / VMEMPAGESIZE); }

    static long getMaxAllocation(void)
        {   return (long(nMaxAllocatedPages) * VMEMPAGESIZE / 1024L); }

    inline BOOL valid(void);

    // This returns the size of the LRU buffer
    inline int getBufSize(void);

    // Returns 0 if not swapping, 1 if swapping to additional memory,
    // or 2 if swapping to additional memory and disk (or disk only).
    inline int isSwapping(void);

    inline long getTextSize(void);

    inline BOOL isShared(void);
    inline const CVMEditor *CurrentOwner(void);

#if defined(CVME_DIAG)
    inline long getLRUCnt(void);
#endif

// ****************************************************************************
    // These control ownership of the buffer
    BOOL addOwner(CVMEditor *anOwner);
    BOOL changeOwner(CVMEditor *anOwner);
    BOOL removeOwner(CVMEditor *anOwner);

    // Retrieve text
    const char *getLine(long line);
    void  getLineLRU(long line);

    // Store text
    void  insertText(long line, char *text);
    void  putLineLRU(void);

    // Delete text
    void  deleteLineInfo(long line, long count, BOOL AllowUndo = TRUE);

    // Undo/redo an operation
    long  UndoRedo(BOOL isUndo);
    void  UndoRedoAvailable(BOOL &bUndo, BOOL &bRedo);

private:
    // Text element position tracking functions
    void  getLineInfo(long line, int *pnPage, int *pnOffset);
    void  setLineInfo(long line, int nPage, int nOffset, BOOL inserting);
    void  getInsPos(int nSize, int *pnPage, int *pnOffset);

    // Undo/redo and free space tracking functions
    void addUndo(long currLineNbr, long newLineNbr, UndoType uType);
    BOOL getFreeSpaceFromUndo(int size, int *pnPage, int *pnOffset);

    // Page management functions
    int getFreeRAMPgNo(void);
    int getFreeVirtPgNo(void);
    int SwapPageIn(int nPage);
    void  SwapPageOut(int nSlot);

#ifdef CVME_DIAG
    void  MarkObsolete(int nPage);
#else
    inline void MarkObsolete(int nPage);
#endif

    // Low level virtual memory functions.  If available memory runs out
    // or if the maximum memory limit is reached, then it starts swapping
    // to disk.
    void initVirtualMemory(void);
    void shutdownVirtualMemory(void);
    void readPage(int nPage, int nSlot);
    void writePage(int nPage, int nSlot);
    void resetVirtualMemory(void);
};

// Inline function bodies
inline BOOL CVirtualMemory::valid(void)
    {   return  bIsValid;    }

// This returns the size of the LRU buffer
inline int CVirtualMemory::getBufSize(void)
    {   return  nBufSize;    }

// Returns 0 if not swapping, 1 if swapping to additional memory,
// or 2 if swapping to additional memory and disk (or disk only).
inline int CVirtualMemory::isSwapping(void)
    {   return (hPageFile != INVALID_HANDLE_VALUE) ? 2 :
               ((nMemPagesAllocated == 0) ? 0 : 1); }

inline long CVirtualMemory::getTextSize(void)
{
    putLineLRU();
    return lTextSize;
}

inline BOOL CVirtualMemory::isShared(void)
    {   return  BOOL(Owners->nextOwner != NULL); }

inline const CVMEditor *CVirtualMemory::CurrentOwner(void)
    {   return  owner; }

#if defined(CVME_DIAG)

// For checking swapping load
inline long CVirtualMemory::getLRUCnt(void)
    {   return lLRU_Cntr;   }

#else

// Fast version for production build
inline void CVirtualMemory::MarkObsolete(int nPage)
{
    int nIdx = SlotLocn[nPage];

    if(SwapSlotNo[nIdx] < 0)
    {
        SwapSlotInUse[-1 - SwapSlotNo[nIdx]] = FALSE;
        SwapSlotNo[nIdx] = 0;
    }
}
#endif

#endif  // __CVIRTMEM_H
