//=============================================================================
// System  : CVMEditor - Virtual Memory Text Editor Library
// File    : SignsOfLife.cpp
// Author  : Eric Woodruff  (Eric_Woodruff@CompuServe.com)
// Updated : Wed 07/08/98 01:00:00
// Note    : Copyright 1997-98, Eric Woodruff, All rights reserved
// Compiler: Visual C++
//
// This file contains an example of the "signs of life" functions that
// can be implemented with the editor.
//
// Version    Date    Who   Comments
// ============================================================================
// 1.0.0    06/16/97  EFW   Ported the code to Visual C++
//=============================================================================

#include "stdafx.h"

#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "CVMEditor.h"
#include "CVMEResource.h"

#include "Resource.h"
#include "ProgDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//=============================================================================

// This is the function that handles the various Signs of Life states.
// 'slLifeSign' indicates what the function is being called for.  'nCmnd'
// indicates which command is being executed.  'lCount' and 'lTotal' represent
// lines done and the total number of lines to do or, for reading text, bytes
// read so far and the total number of bytes to be read.  Note that if either
// GlobalOptions.initThreshold or GlobalOptions.reportInterval is zero, then
// the editor classes will not issue signs of life.  This function can be as
// simple or as sophisticated as you want it to be.
//
BOOL doSignsOfLife(SignsOfLife slLifeSign, UINT nCmnd, long lCount,
    long lTotal)
{
    static CProgressDlg *progDlg = NULL;
    static char *pszOperation;
    static int nSteps;

    BOOL bRetVal = TRUE;
    char szText[256];

    // Ignore it if the progress dialog couldn't be created
    if(!progDlg && slLifeSign != slInitiate)
        return TRUE;

    switch(slLifeSign)
    {
        case slInitiate:
            progDlg = new CProgressDlg;
            if(!progDlg)
                break;

            progDlg->Create(NULL);


            // These are all the commands that issue signs of life
            switch(nCmnd)
            {
                case CVM_MOVEBLOCK:
                    pszOperation = "Mov";
                    break;

                case CVM_COPYBLOCK:      // Including clipboard operations
                    pszOperation = "Copy";
                    break;

                case CVM_WRAPPARA:
                    pszOperation = "Reformat";
                    break;

                case CVM_READFILE:       // Reports progress in bytes not lines
                    pszOperation = "Read";
                    break;

                case CVM_WRITEBLOCK:
                    pszOperation = "Writ";
                    break;

                // NOTE: If you see delete starting a clipboard copy or cut
                //       operation, don't panic!  It's just the old block
                //       of text being cleared from the clipboard.
                case CVM_EDITCLEAR:
                    pszOperation = "Delet";
                    break;

                case CVM_EDITUNDO:   // Serves both undo and redo operations
                    pszOperation = "Undoing/Redo";
                    break;

                case CVM_EDITFIND:
                case CVM_FINDMATCHING:
                    pszOperation = "Search";
                    break;

                default:    // All other commands (from alterText())
                    pszOperation = "Alter";
                    break;
            }

            // Note: If the reporting interval is small and the total
            // dividided by it is larger than USHRT_MAX, the progress bar
            // will wrap.
            nSteps = int(lTotal / CVMEditor::GlobalOpts.nReportInterval);

            progDlg->SetRange(0, nSteps);
            progDlg->SetPos(int(lCount /
                CVMEditor::GlobalOpts.nReportInterval));
            // Fall through and display the message to complete the case.

        case slReporting:
        case slDone:
            sprintf(szText, "Please Wait -- %sing %ld of %ld",
                pszOperation, lCount, lTotal);

            progDlg->SetStatus(szText);
            progDlg->SetPos(int(lCount /
                CVMEditor::GlobalOpts.nReportInterval));

            if(slLifeSign == slDone)
            {
                delete progDlg;
                progDlg = NULL;      // Reset pointer ready for next time.
            }
            else
                bRetVal = !progDlg->CheckCancelButton();
            break;

        // Unused by the default editor classes,
        // but they are here incase you need them.
//        case slHide:
//            progDlg->ShowWindow(SW_HIDE);
//            break;
//
//        case slShow:
//            progDlg->ShowWindow(SW_SHOW);
//            break;
    }

    return bRetVal;
}
